#!/usr/bin/python
'''Test a codebase (principally CPMD/NECI) against existing results.
   James Spencer

See README for more information and how to test other codes.'''

from optparse import OptionParser,OptionValueError
import commands,glob,re,shutil,sys,time,os,os.path,signal,threading

# Version refers to the svn version number in the wwmm repository.
__version__='10684'

# --- Legacy compatibility ---

# Using subprocess for the (possibly lengthy) system calls to run 
# the calculations allows for signal capture (e.g. ctrl-c!), but the 
# crippled UCL machines have an old version of python installed w/out 
# subprocess, so use os.system as a fallback.
try:
    import subprocess
except ImportError:
    print "Import subprocess failed.  Using os.system."
# Another workaround for the crippled UCL machines
try:
    set(['abc'])
except NameError:
    from sets import Set as set
# If using python <2.5, all doesn't exist. Define our own.
try:
    all([])
except NameError:
    def all(iterable):
        for e in iterable:
            if not e:
                return False
        return True

# --- Error handling.  ---

def SignalHandler(signal,frame):
    '''Capture interupt signal and leave quietly.'''
    print 'Ctrl-C has been caught.  Bye!'
    sys.exit(1)


class RunError(Exception):
    pass

# --- Option parsing.  Routines also initialise and set up test runs.  ---

def check_tidy(option,opt_str,value,parser):
    '''Callback function for testing the arguments to --tidy flag.'''
    if value < 0:
        raise OptionValueError("%s option takes a positive value" % opt_str)
    setattr(parser.values,option.dest,value)


def ParseOptions(my_args):
    '''Parse command line options using optparse.'''
    usgmsg='''./testcode.py [options]
Test code against previous results.  Requires extractdata.pl.
Default: test all jobs in the shortjobs category.
See README for more details.'''

    parser=OptionParser(usage=usgmsg)
    
    parser.add_option('-b','--benchmark',help='Specify a different benchmark file by revision number. See README for more details.')
    parser.add_option('-c','--compare',action='store_true',default=False,help='Compare old test file (specified via -r) to benchmark.')
    parser.add_option('-e','--exe',default=[],action='append',help='Set executable locations, via either an alternative key stored in userconfig or using the syntax jobtype:executable_location.  See README for further details.')
    parser.add_option('-i','--inputregex',default='', help='Use input files in each job directory which match the provided regular expression.')
    parser.add_option('-j','--jobcat',default=[],action='append',help='Run tests on jobs in JOBCAT.  Can be specified multiple times.')
    parser.add_option('--jr','--jobregex',dest='jobregex',default='',help='Run tests on all jobs that match JOBREGEX.')
    parser.add_option('-p','--printcats',action='store_true',default=False,help='Print available job categories.')
    parser.add_option('-q','--queue',action='store_true',default=False,help='Run the tests by submitting them to a cluster queue.  See README for more details.')
    parser.add_option('-r','--revno',default='',help='Choose a suffix for test output. Default: generate a unique date-based filename. See README for more details.')
    parser.add_option('--runbench',action='store_true',default=False,help='Run jobs and produce a new set of benchmarks.')
    parser.add_option('-t','--tol', type='float',default=1.e-10,help='Tolerance for testing agreement between test jobs and benchmarks.  Default=%default.')
    parser.add_option('--tidy',action="callback",callback=check_tidy,type='int',dest='ndays',default=-1,help='Remove test.out*, submit.sh* and testcode_job* files older than NDAYS from each test directory.')
    
    (options,args)=parser.parse_args(my_args)
    if len(args)>0:
        print 'Do not understand arguments specified: ',args,'.'
        parser.print_usage()
        sys.exit(1)
    else:
        return options


def ParseExeOptions(exeoptions,jobtypes):
    '''Search for the job type(s) supplied in the argument list and set the executable accordingly. Executable option can be set in 3 different ways:
    1. KEY
       Set all executables to use the KEY executable as given in userconfig.py.
    2. jobtype:KEY
       Same as 1, but only for the specified jobtype.
    3. jobtype:executable_location
       Set jobtype executable to be the specified location.
jobtype['exe'] is set to be the executable provided for all jobtype given in the jobtypes list.'''
    for jobopt in exeoptions:
        if ':' in jobopt:
            [jobtype,jobexe]=jobopt.split(':')
            try:
                # 2. jobtype:KEY
                jobtypes[jobtype]['exe']=jobtypes[jobtype][jobexe]
            except KeyError:
                # 3. jobtype:executable_location
                jobexe=os.path.realpath(os.path.expanduser(jobexe))
                if jobtype in jobtypes:
                    jobtypes[jobtype]['exe']=jobexe
                else:
                    print 'Do not understand executable option: %s.' % jobopt
                    sys.exit(1)
                if not os.path.exists(jobexe):
                    print 'Given location: %s in executable option: %s does not exist.'% (jobexe,jobopt)
                    sys.exit(1)
        else:
            for jobtype in jobtypes:
                try:
                    # 1. KEY.
                    jobtypes[jobtype]['exe']=jobtypes[jobtype][jobopt]
                except KeyError:
                    print 'Do not understand executable option: %s.' % jobopt
                    sys.exit(1)


def GetJobList():
    '''List jobs to do.

The tests run are actually split up by two categorisations:
    1. The directories (the "jobs") as given in jobconfig.py, selected via -j and --jr.
    2. The various input files within each job directory, which are found by globbing the
    job type input filename in userconfig.py.  These can be further selected via a regular
    expression using the -i flag.
Returns the list of directories of the test jobs (the first categorisation) and compareall.
The second categorisation is evaluated as each job set is run.
compareall is true if we are comparing existing output and no jobs are provided on the command line

usage: (jobs,compareall)=GetJobList()
Accesses options and jobcategories from the global namespace.'''
    global options,jobcategories
    jobs=[]
    for cat in options.jobcat:
        if cat in jobcategories:
            jobs+=jobcategories[cat]
        else:
            print 'Job category: %s does not exist.' % cat
            sys.exit(1)
    if options.jobregex!='':
        testmatch=False
        for cat in jobcategories:
            for job in jobcategories[cat]:
                if re.search(options.jobregex,job['dir']) and job not in jobs:
                    testmatch=True
                    jobs.append(job)
        if not testmatch:
            print 'No jobs match %s.' % options.jobregex
            sys.exit(1)
    # If we're comparing, and no jobs are provided on the command line, then we'll compare 
    # every job for which the test file exists (check for this later).
    if not jobs and options.compare:
        compareall=True # Use this to suppress output if test file doesn't exist.
        for cat in jobcategories:
            for job in jobcategories[cat]:
                jobs.append(job)
    else:
        compareall=False
    if not jobs:
        # List of jobs is still empty.  Run defaultjobs category.
        jobs=defaultjobs
    if options.runbench:
        if not jobs:
            print 'Running tests on alljobs category in addition to those specified.'
        # Produce new benchmarks for alljobs plus any provided by command line arguments.
        jobs+=jobcategories['alljobs']
        if not options.inputregex:
            print 'Running tests on all input files.  Resetting inputregex option.'
            options.inputregex=''
    # Remove duplicate jobs from list.  This is O(N^2), whereas a more efficient 
    # implementation is around O(N).  N should be quite small though, so we don't 
    # need Elaine Roberts (www.xkcd.com/342/).
    tmpjobs=[]
    for job in jobs:
        if job not in tmpjobs:
            tmpjobs.append(job)
    jobs=tmpjobs
    return (jobs,compareall)


def ParseFilename(file_id,type='benchmark'):
    '''Return the filename based on the file_id.

    If type=test, return test%(INPUTTAG)s.out-file_id.

    If file_id=b:file_id, override the type argument and return the benchmark
    filename.
    If file_id=t:file_id, override the type argument and return the test
    filename.'''

    if ':' in file_id:
        id_list=file_id.split(':')
        if len(id_list)!=2:
            print "Error parsing file option %s." % file_id
            sys.exit(1)
        if id_list[0]=='t':
            file_id=id_list[1]
            type='test'
        elif id_list[0]=='b':
            file_id=id_list[1]
            type='benchmark'
    if type=='benchmark':
        return 'benchmark%s-%s.out' % ('%(INPUTTAG)s',file_id)
    elif type=='test':
        return 'test%s.out-%s' % ('%(INPUTTAG)s',file_id)
    else:
        print "Error parsing file option: bad type %s." % type
        sys.exit(1)


def GetTBFilenames(jobs,datefmt='%d%m%y'):
    '''Return the benchmark and test filenames.

The benchmark filename is specified in userconfig unless provided via the options.    
The test filename is unique for the current set of job tasks based on the date (using 
the date format, default ddmmyy, provided) unless specified via the options.'''
    global options,benchfile
    # benchmark filename.
    if options.benchmark:
        # User has specified a different benchfile.
        benchfile=ParseFilename(options.benchmark,'benchmark')
    elif 'INPUTTAG' not in benchfile:
        # For backwards compatibility.
        benchfile=benchfile.replace('benchmark','benchmark%(INPUTTAG)s')

    # filename for test output.
    if options.revno!='':
        testfile=ParseFilename(options.revno,'test')
    else:
        # Get unique filename for test output if not specified.
        testfile='test%(INPUTTAG)s.out-'+time.strftime('%d%m%y')
        i=0
        for job in jobs:
            os.chdir(job['dir'])
            for jobtag,jobinpfile in lsjobinps(job['type']['inp']):
                jobtestfile=testfile % {'INPUTTAG':jobtag}
                if os.path.exists(os.path.join(jobtestfile)):
                    i=max(i,1)
                    while os.path.exists(jobtestfile+'-'+str(i)):
                        i+=1 
            os.chdir(testcodedir)
        if i>0:
            testfile=testfile+'-'+str(i)
    return (benchfile,testfile)


def lsjobinps(jobtypeinp):
    '''List all the inputs in the current directory matching jobtypeinp
    and return the next inputfile we need to consider and its "tag" (the unique
    part of the input file name).'''
    if '*' in jobtypeinp:
        inpprefix,inpsuffix=jobtypeinp.split('*')
    else:
        inpprefix,inpsuffix=jobtypeinp,''
    for jobinpfile in glob.glob(jobtypeinp):
        if re.search(options.inputregex,jobinpfile):
            # Find the tag for this input file matched by the * in the job type input file pattern.
            jobtag=re.search('%s(.*)%s' % (inpprefix,inpsuffix),jobinpfile).groups()[0]
            yield jobtag,jobinpfile

# --- Data extraction utilities: used in comparing test data with benchmark data. ---

def IsNaN(x):
    '''Test for x being NaN.  This is not guaranteed to be portable, but does work under Linux.'''
    return type(x) is float and x!=x

def TryStrToFloat(x):
    '''If possible, convert x from a string to a float, otherwise leave x unchanged.'''
    try:
        x=float(x)
    except ValueError:
        pass
    return x


def DictsFromTabular(tabdat):
    '''Convert an set of tabular data in a string to a list of dicts of tuples.  Where possible, convert entries to floats.

    The first row and any subsequent rows containing no numbers are assumed to form headers of a subtable, and so form the keys for the subsequent subtable.

e.g. a  b  c   ->   [{'a':(1,4),'b':(2,5),'c':(3,6)}]
     1  2  3 
     4  5  6 
and  
     a  b  c   ->   [{'a':(1,4),'b':(2,5),'c':(3,6)},{'a':(7),'b':(8),'d'(9)}] 
     1  2  3 
     4  5  6 
     a  b  d 
     7  8  9 
'''
    data=[i.split() for i in tabdat.splitlines()]
    data=[map(TryStrToFloat,data[i]) for i in range(len(data))]
    data2=[[data[0]]]
    for d in data[1:]:
        if False in [isinstance(i,str) for i in d]:
            data2[-1].append(d)
        else:
            data2.append([d])
    data=[ dict(zip(d[0],zip(*d[1:]))) for d in data2 ]
    return data


def PrintData(datatype,datadicts,dataout,printheader):
    '''Print output from extraction script.  Prepend non-header lines
with the datatype string.  Print header lines if printheader=True.'''
    headerindices=[0]
    for i in range(1,len(datadicts)):
        headerindices.append(i+len(datadicts[i-1].values()[0]))
    datalines=dataout.splitlines()
    for i in range(len(datalines)):
        if i in headerindices:
            if printheader:
                print '%-10s %s' % ('',datalines[i])
        else:
            print '%-10s %s' % (datatype,datalines[i])

# --- Cluster run routines. ---

class JobThread(threading.Thread):
    '''Thread containing a job to be submitted to a cluster queueing system.'''
    def __init__(self,job,jobinpfile,jobtestfile,subfile,jobbenchfile):
        self.job=job
        self.jobinpfile=jobinpfile
        self.jobtestfile=jobtestfile
        self.subfile=subfile
        self.jobbenchfile=jobbenchfile
        self.passed=False
        threading.Thread.__init__(self)
    def run(self):
        '''Run and analyse a test job.'''
        # Submit job to the queue.
        lock.acquire()
        # Other threads have already changed directory on us!
        os.chdir(testcodedir)
        os.chdir(self.job['dir'])
        print 'Submitting test using %s (input file) and %s (template submit file) in %s.' % (self.jobinpfile,self.job['type']['submit_template'],self.job['dir'])
        # Nest the try/finally block for 2.4 compatibility.
        # In 2.5 this becomes try: except: finally:.
        try:
            try:
                jobid=RunJob(self.job,self.jobinpfile,self.jobtestfile,self.subfile)
            except RunError:
                self.passed=False
                return
        finally:
            lock.release()
        # Wait for it to complete.
        WaitForQJob(jobid)
        # Analyse output.
        #  1. Stop other threads from screwing us over.
        #     If we don't lock, then other threads can change directory, write 
        #     over the benchmark/test data, etc.
        lock.acquire()
        #  2. Other threads have already changed directory on us!
        os.chdir(testcodedir)
        os.chdir(self.job['dir'])
        #  3. Compare.
        print '\nRunning test using %s in %s.' % (self.jobinpfile,self.job['dir'])
        try:
            try:
                self.passed=CompareTest(self.job,self.jobinpfile,self.jobtestfile,self.jobbenchfile)
            except RunError:
                self.passed=False
        finally:
        #  4. Allow other threads to resume.
            lock.release()
    

def WaitForQJob(jobid):
    '''Call qstat every minute to see if jobid has finished.'''
    retcode=0
    while retcode==0:
        time.sleep(60)
        qstat=subprocess.Popen('qstat %s' % jobid,shell=True,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
        qstat.wait()
        # If the return code to qstat is non-zero, then the jobid no longer 
        # exists: must have ended, one way or the other!
        retcode=qstat.returncode


def CreateSubFile(job,jobinpfile,jobtestfile,subfile,runcmd):
    '''Create a submit file for sending test job to the queue.  This is a dumb routine: 
it only modifies a template file in the testcode directory.'''
    shutil.copy(job['type']['submit_template'],subfile)
    f=open(subfile,'r+')
    subcontents=f.read()
    subcontents=subcontents.replace('testcode.jobdir',job['dir'])
    subcontents=subcontents.replace('testcode.jobinput',jobinpfile)
    subcontents=subcontents.replace('testcode.joboutput',jobtestfile)
    subcontents=subcontents.replace('testcode.runcmd',runcmd) 
    f.seek(0)
    f.write(subcontents)
    f.close()

# --- Main run routines. ---

def TidyUp():
    '''Clean up: delete all test.out* files older than ndays.'''
    print 'Deleting all test.out-*,submit.sh* and testcode_job* files older than %s days from each job directory.' % (options.ndays)
    tmpdirs,dirs=[],[]
    remove_files=['test.out-*','submit.sh*','testcode_job*']
    for cat in jobcategories:
        for job in jobcategories[cat]:
            tmpdirs.append(job['dir'])
    for dir in tmpdirs:
        if dir not in dirs:
            dirs.append(dir)
    then=time.time()-options.ndays*86400
    for dir in dirs:
        os.chdir(dir)
        for file in glob.glob('test.out-*'):
            if os.stat(file)[-2] < then:
                os.remove(file)
        os.chdir(testcodedir)
    sys.exit()


def RunJob(job,jobinpfile,jobtestfile,subfile):
    '''Run the jobinpfile belonging to the job test(s) dictionary.  Output to jobtestfile.  
If submitting job to a (PBS/Torque) queueing system, then return the jobid.'''
    global options
    if not os.path.exists(job['type']['exe']):
        alljobs=False
        print 'Executable: %s does not exist.' % (job['type']['exe'])
        raise RunError
    if os.path.exists(jobinpfile):

        # This is a little hack-y.  It would be more pythonic to use subprocess to 
        # capture the stdoutput etc, but this way allows us to follow tests interactively...
        # It also makes submitting test jobs to a queueing system very easy.
        runcmd='%s %s > %s 2> %s' % ( job['type']['exe'],jobinpfile,jobtestfile,'testcode.err' )
        if options.queue:
            CreateSubFile(job,jobinpfile,jobtestfile,subfile,runcmd)
            runcmd='qsub %s' % subfile
            p=subprocess.Popen(runcmd,shell=True,stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
            p.wait()
            jobid=p.communicate()[0].strip()
            return jobid
        else:
            try:
                retcode=subprocess.call(runcmd,shell=True)
            except NameError:
                retcode=os.system(runcmd)
                if retcode!=0:
                    print 'Calculation was terminated by signal: ',retcode
                    sys.exit(1)
            except OSError, detail:
                print 'Executation of test using %s in %s failed: ' % (jobinpfile,job['dir']),detail

    else:
        print 'Input file: %s in %s does not exist.' % (jobinpfile,job['dir'])
        raise RunError


def CompareTest(job,jobinpfile,jobtestfile,jobbenchfile):
    '''Compare test output to benchmark.
    
This assumes output in a (space-separated) tabular format, 
where the first line is the column headings and subsequent lines are
data (or an extension thereof, as specified in DictsFromTabular).'''
    global testcodedir,jobtimecols
    jobpassed=True
    printbench=True
    printtestheader=False

    if os.path.exists(jobtestfile):
        # Use commands as it's easier to get output and extraction should be fast: don't need the 
        # error capturing provided by subprocess.
        testout=commands.getoutput(job['type']['extract']+' '+jobtestfile)
    else:
        print 'Test file: %s does not exist.' % jobtestfile
        jobpassed=False
        raise RunError
    if os.path.exists(jobbenchfile):
        benchout=commands.getoutput(job['type']['extract']+' '+jobbenchfile)
    else:
        print 'Benchmark: %s does not exist.' % jobbenchfile
        jobpassed=False
        printbench=False
        printtestheader=True

    # Convert data into dictionaries.  Assumptions about the format of the data are
    # made in DictsFromTabular.
    testdatadicts=DictsFromTabular(testout)
    if printbench:
        benchdatadicts=DictsFromTabular(benchout)

    if printbench: # We have a benchfile to compare to.
        if len(testout.splitlines())!=len(benchout.splitlines()):
            jobpassed=False
            printtestheader=True
            print 'Different amounts of data extracted from test output and benchmark.'
        if len(testdatadicts)!=len(benchdatadicts):
            jobpassed=False
            printtestheader=True
            print 'Different amounts of data extracted from test output and benchmark.'
        else:
            for i in range(len(benchdatadicts)):
                benchdata=benchdatadicts[i]
                testdata=testdatadicts[i]
            # Check headers are the same (ie at least the same type of calculation is done!).
            benchkeys=set(benchdata.keys())
            testkeys=set(testdata.keys())
            if benchkeys.symmetric_difference(testkeys):
                # Keys are not the same: something's gone wrong.
                jobpassed=False
                printtestheader=True
                print 'Different amounts of data extracted from test output and benchmark.'
            else:
                try: 
                    # Test each non-timing column of the output.
                    for k in benchkeys.difference(set(jobtimecols)):

                        # See if the default tolerance has been overridden for this data value for this job.
                        tolerance_key='%s %s tolerance' % (jobinpfile,k)
                        if job.has_key(tolerance_key):
                            tolerance=job[tolerance_key]
                            print 'Testing %s with a non-default tolerance: %#-5.4g.' % (k,tolerance)
                        else:
                            tolerance=options.tol

                        # Test each row of the current column.
                        for i in range(len(benchdata[k])):
                            # Test to make sure floats are actually numbers!
                            if IsNaN(benchdata[k][i]) or IsNaN(testdata[k][i]):
                                print 'Cannot compare NaNs.  Failing %s.' % (k,)
                                jobpassed=False
                            elif isinstance(benchdata[k][i],float):
                                if abs(benchdata[k][i]-testdata[k][i]) > tolerance:
                                    print '%s for benchmark and test differ by %#-5.4g.' % (k,benchdata[k][i]-testdata[k][i])
                                    jobpassed=False
                                elif tolerance is not options.tol:
                                    print '%s for benchmark and test agree to within %#-5.4g.' % (k,benchdata[k][i]-testdata[k][i])
                            else:
                                # If the data item isn't a float, require it to be equal (as defined by python).
                                if benchdata[k][i]!=testdata[k][i]:
                                    print '%s for benchmark and test differ.' % (k,)
                                    jobpassed=False
                except Exception,err:
                    print 'Error comparing test to benchmark.',err
                    jobpassed=False
                    printtestheader=True

    # For output, let the user's data extraction program do the formatting.
    if printbench:
        if jobpassed:
            print 'Test: passed.'
        else:
            print 'Test: **FAILED**.'
        PrintData('Benchmark',benchdatadicts,benchout,True)
    # To do: fix pretty printing.  Cases:
    # 1) # of dicts don't match.  Print benchmark and test output separately.
    # 2) # of dicts match.  Rows in dicts don't.  Interleave bench & test.  Print test headers.
    # 3) # dict keys don't match.  Interleave bench & test.  Print test headers.
    # 4) # Tests pass.    Interleave bench & test.  Do not print test headers.
    if printtestheader:
        PrintData('Test',testdatadicts,testout,True)
    else:
        PrintData('Test',testdatadicts,testout,False)
    return jobpassed


def MakeBenchfiles(allpassed,starttime,jobtypes,jobs,testfile):
    '''If the tests have passsed (or user overrides), make the set of testjobs the new benchmarks.'''
    global options
    if not allpassed:
        print 'Warning: failed tests or problems running tests.'
        ans=''
        while ans!='y' and ans!='n':
            ans=raw_input('Set tests to be new benchmark? [y/n]')
            
    if allpassed or ans=='y':
        # Get highest svn revision number from executable paths.
        # Assume codes that are tested together have are sufficiently intertwined such that 
        # taking the maximum revision number for all codes is the sane choice.
        coderevision=0
        for jobtype in jobtypes:
            exe=jobtypes[jobtype]['exe']
            exerevision=int(commands.getoutput('svn info '+os.path.dirname(exe)+' 2> /dev/null | grep "Last Changed Rev"').split()[-1])
            coderevision=max(coderevision,exerevision)
        if coderevision==0:
            coderevision=int(raw_input('Source(s) not under svn control.  Enter new benchmark revision number. '))
        newbenchfile='benchmark%(INPUTTAG)s-r'+str(coderevision)+'.out'
        print 'Setting tests to be new benchfiles, ',newbenchfile+'.'
        
        for job in jobs:
            os.chdir(job['dir'])
            for jobtag,jobinpfile in lsjobinps(job['type']['inp']):
                # Find the tag for this input file matched by the * in the job type input file pattern.
                jobtestfile=testfile % {'INPUTTAG':jobtag}
                newjobbenchfile=newbenchfile % {'INPUTTAG':jobtag}
                # Copy testfile to benchmark output.
                try:
                    shutil.copy(jobtestfile,newjobbenchfile)
                except:
                    print 'Testfile %s does not exist in %s.'% (jobtestfile,job['dir'])
                    print 'Not updating benchmarks.'
                    sys.exit(1)
                # Copy all additional files created by the programs to 
                # a benchfiles subdirectory.
                if not os.path.isdir('benchfiles'):
                    os.mkdir('benchfiles')
                for file in glob.glob('*'):
                    lastmodtime=os.stat(file)[-2]
                    if lastmodtime > starttime and os.path.isfile(file) and not re.match('testcode.err|test.out-',file):
                        shutil.copy(file,'benchfiles/')
            os.chdir(testcodedir)
        if options.compare:
            print 'Warning: not copying files to benchfiles folder.'
            
        # Change benchmark in config module.
        f=open('userconfig.py','r')
        s=f.read().replace(benchfile,newbenchfile)
        s=s.replace(benchfile%{'INPUTTAG':''},newbenchfile) # This is for the format of the old benchfile, where multiple tests within one directory were not allowed.
        f=open('userconfig.py','w')
        f.write(s)
        f.close()
        
# --- Omnipotent power. ---

def DoTestCode():
    '''Wrap the guts of testcode in a function.  This makes it easy to use testcode as a module.'''
    # We can access a variable from outside the local namespace, but must 
    # define it to be global in order to change it.  To not declare all such 
    # variables global is a dubious practice, but one we shall use to allow 
    # freedom in the config files.
    # from userconfig
    global benchfile,jobtypes,jobtimecols
    # from jobconfig
    global jobcategories
    # from testcode
    global options

    if options.printcats:
        print 'Available job categories:'
        for category in jobcategories:
            print category
        sys.exit()

    # Set executable locations if specified on cli.
    ParseExeOptions(options.exe,jobtypes)

    # Build up list of jobs to do.
    (jobs,compareall)=GetJobList()

    # See if user has set the keys for the timing columns in the output.
    try:
        t=jobtimecols
    except NameError:
        jobtimecols=[]
    try:
        datefmt=my_date_fmt
    except NameError:
        # default date format
        datefmt='%d%m%y'

    (benchfile,testfile)=GetTBFilenames(jobs,datefmt)

    # Print out useful info before testing.
    for jobname in jobtypes:
        # Take this opportunity to follow relative paths and expand ~ (might have already done so).
        jobtypes[jobname]['exe']=os.path.realpath(os.path.expanduser(jobtypes[jobname]['exe']))
        print '%-20s %s.' % ('%s executable:' % jobname,jobtypes[jobname]['exe'])
    print '%-20s %s.' % ('Benchmarks: ',(benchfile) % {'INPUTTAG':'*'})
    print '%-20s %s.' % ('Test outputs: ',(testfile) % {'INPUTTAG':'*'})
    print '%-20s %#-5.4g.' % ('Tolerance: ',options.tol)
    if abs(options.tol)>1.e-5:
        print 'Tolerance is worryingly high.  Are you sure?'
    if options.queue:
        print '' # Anal-retentive.

    # Run jobs and compare to existing results.
    passed=[]
    jobthreads=[]
    starttime=time.time()
    for job in jobs:

        try:
            os.chdir(testcodedir)
            os.chdir(job['dir'])
        except:
            print '\nFailed to move to directory: '+job['dir']+'.'
            passed.append(False)
            continue

        joblist=lsjobinps(job['type']['inp'])
        try:
            joblist.next()
        except StopIteration:
            print '\n**FAILED**.  No job input files exist in directory: %s.' % job['dir']
            passed.append(False)
            os.chdir(testcodedir)
            continue

        # For each input file in this job category.
        for jobtag,jobinpfile in lsjobinps(job['type']['inp']):

            jobtestfile=testfile % {'INPUTTAG':jobtag}
            jobbenchfile=benchfile % {'INPUTTAG':jobtag}
            jobsubfile='submit%(INPUTTAG)s.sh' % {'INPUTTAG':jobtag}

            if options.compare and compareall and not os.path.exists(jobtestfile):
                # Silently go to next job.
                continue
            elif options.compare:
                print '\nComparing test using %s in: %s.'% (jobinpfile,job['dir'])
            else:
                # Submitting jobs to the queue on a cluster.
                # Set each test job in its own thread.
                # Create all threads first and then start them to avoid
                # race conditions and lock issues.
                if options.queue:
                    jobthreads.append(JobThread(job,jobinpfile,jobtestfile,jobsubfile,jobbenchfile))
                else:
                    print '\nRunning test using %s in %s.' % (jobinpfile,job['dir'])
                    try:
                        RunJob(job,jobinpfile,jobtestfile,jobsubfile)
                    except RunError:
                        passed.append(False)
                        os.chdir(testcodedir)
                        continue

            # Compare test to benchmark.  If submitted to a queue, the thread does the comparison.
            if not options.queue:
                try:
                    passed.append(CompareTest(job,jobinpfile,jobtestfile,jobbenchfile))
                except RunError:
                    passed.append(False)
                    os.chdir(testcodedir)
                    continue

        # Have done all the test jobs in the current directory.
        os.chdir(testcodedir)

    for t in jobthreads:
        # Start threads if queueing.
        t.start()
    for t in jobthreads:
        # Wait for all threads to end.
        t.join()
    if options.queue:
        # See if all jobs passed!
        passed.extend([job.passed for job in jobthreads])

    allpassed=all(passed)
    if options.runbench:
        MakeBenchfiles(allpassed,starttime,jobtypes,jobs,testfile)
    elif allpassed:
        print 'Tests successfully passed.'
    elif not allpassed:
        print 'Warning: failed tests or problems running tests.'

    # Clean up: I don't like keeping the compiled modules around, and the time saving is minimal.
    os.chdir(testcodedir)
    for localmodule in glob.glob('*pyc'):
        os.remove(localmodule)

    return allpassed

# --- Initialisation. ---

allpassed=False # Assume the worst, initially!
lock=threading.Lock()
testcodedir=os.getcwd()
signal.signal(signal.SIGINT,SignalHandler) # Listen out for Ctrl-C.  

if __name__=='__main__':
    # Import user data here so that the help options can be seen without setting up userconfig.py.
    options=ParseOptions(sys.argv[1:])
    try:
        from jobconfig import *
    except ImportError,detail:
        print "Error.",detail,'.'
        print "Try creating a local userconfig.py file. See README for details or run setup.py."
        sys.exit(1)
    if options.ndays > -1:
        # Not actually testing code; rather tidying up.
        TidyUp()
    else:
        DoTestCode()
else:
    options=ParseOptions([])
    try:
        from jobconfig import *
    except ImportError,detail:
        print "Error.",detail,'.'
        print "Try creating a local userconfig.py file. See README for details or run setup.py."
        sys.exit(1)
